package b4a.flm.overlaywdw;

import android.graphics.PixelFormat;
import android.graphics.drawable.Drawable;
import android.os.Build;
import android.view.Gravity;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.View.OnLongClickListener;
import android.view.WindowManager.LayoutParams;
import anywheresoftware.b4a.BA;
import anywheresoftware.b4a.objects.ButtonWrapper;
import anywheresoftware.b4a.objects.drawable.StateListDrawable;

@BA.ShortName("OverlayButton")
@BA.Events(values = {
		"Click", "LongClick"
})
public class OverlayButton implements OverlayObject
{
	ButtonWrapper Btn;
	LayoutParams LPBtn;
	boolean VueAjoutef;

	/**
	 *Creates a button at the specified position. The button is displayed after calling Add.
	 *The button cannot be placed or moved over the status and notification bars.
	 *@param Width Width of the button.
	 *@param Height Height of the button.
	 *@param EventPrefix Prefix of the Click and LongClick events.
	 */
	public void Initialize(final BA ba, int X, int Y, int Width, int Height, final String EventPrefix)
	{
		InnerInitialize(ba, X, Y, Width, Height, false, EventPrefix);
	}

	/**
	 *Creates a button at the specified position. The button is displayed after calling Add.
	 *The button can be placed or moved over the status and notification bars. Starting from Android 8 (Oreo), it always appears under critical system windows.
	 *@param Width Width of the button.
	 *@param Height Height of the button.
	 *@param EventPrefix Prefix of the Click and LongClick events.
	 */
	public void Initialize2(final BA ba, int X, int Y, int Width, int Height, final String EventPrefix)
	{
		InnerInitialize(ba, X, Y, Width, Height, true, EventPrefix);
	}

	@SuppressWarnings("deprecation")
	private void InnerInitialize(final BA ba, int X, int Y, int Width, int Height, boolean OverBars, final String EventPrefix)
	{
		Btn = new ButtonWrapper();
		Btn.Initialize(ba, "");
		if (EventPrefix.length() > 0) {
			final String Prefixe = EventPrefix.toLowerCase(BA.cul);
			if (ba.subExists(Prefixe + "_click")) {
				Btn.getObject().setOnClickListener(new OnClickListener() {
					@Override
					public void onClick(View Sender) {
						ba.raiseEvent2(Sender, true, Prefixe + "_click", true, new Object[0]);
					}
				});
			}
			if (ba.subExists(Prefixe + "_longclick")) {
				Btn.getObject().setOnLongClickListener(new OnLongClickListener() {
					@Override
					public boolean onLongClick(View Sender) {
						ba.raiseEvent2(Sender, true, Prefixe + "_longclick", true, new Object[0]);
						return true;
					}
				});
			}
		}
		int LP;
		if (OverBars) {
			if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O)
				LP = LayoutParams.TYPE_APPLICATION_OVERLAY;
			else
				LP = LayoutParams.TYPE_SYSTEM_ERROR;
			LPBtn = new LayoutParams(Width, Height, X, Y, LP, LayoutParams.FLAG_NOT_FOCUSABLE | LayoutParams.FLAG_LAYOUT_IN_SCREEN, PixelFormat.TRANSLUCENT);
		}
		else {
			if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O)
				LP = LayoutParams.TYPE_APPLICATION_OVERLAY;
			else
				LP = LayoutParams.TYPE_SYSTEM_ALERT;
			LPBtn = new LayoutParams(Width, Height, X, Y, LP, LayoutParams.FLAG_NOT_FOCUSABLE, PixelFormat.TRANSLUCENT);
		}
		LPBtn.gravity = Gravity.TOP + Gravity.LEFT;
		VueAjoutef = false;
	}

	@Override
	public boolean IsInitialized()
	{
		return (Btn != null && Btn.IsInitialized());
	}

	/** Returns the wrapped button. */
	public ButtonWrapper getButton()
	{
		return Btn;
	}

	/** Associates this button with the given overlay window so the button is automatically added or removed when the window is opened or closed, and automatically shown or hidden according to the window visibility. */
	public void BindTo(OverlayWindow Window)
	{
		if (Window.Bound.indexOf(this) == -1)
			Window.Bound.add(this);
	}

	/** Breaks the association between this button and the given overlay window. */
	public void UnbindFrom(OverlayWindow Window)
	{
		Window.Bound.remove(this);
	}

	/** Adds this button to the screen. */
	public void Add()
	{
		if (!VueAjoutef) {
			OverlayWindow.WM.addView(Btn.getObject(), LPBtn);
			VueAjoutef = true;
		}
	}

	/** Removes this button from the screen. */
	public void Remove()
	{
		if (VueAjoutef) {
			OverlayWindow.WM.removeView(Btn.getObject());
			VueAjoutef = false;
		}
	}

	/** Returns True if the button was added by a call to Add (or a call to Open if bound). */
	public boolean IsAdded()
	{
		return VueAjoutef;
	}

	/** Sets the images for the Up, Down and Disabled states. Pass Null if you do not want to set an image. */
	public void SetStateImages(Drawable Up, Drawable Down, Drawable Disabled)
	{
		StateListDrawable SLD = new StateListDrawable();
		SLD.Initialize();
		if (Down != null)
			SLD.AddState(StateListDrawable.State_Pressed, Down);
		if (Disabled != null)
			SLD.AddState(StateListDrawable.State_Disabled, Disabled);
		if (Up != null)
			SLD.AddCatchAllState(Up);
		Btn.setBackground(SLD.getObject());
	}

	@Override
	public int getX()
	{
		return LPBtn.x;
	}
	@Override
	public void setX(int X)
	{
		LPBtn.x = X;
		OverlayWindow.WM.updateViewLayout(Btn.getObject(), LPBtn);
	}

	@Override
	public int getY()
	{
		return LPBtn.y;
	}
	@Override
	public void setY(int Y)
	{
		LPBtn.y = Y;
		OverlayWindow.WM.updateViewLayout(Btn.getObject(), LPBtn);
	}

	@Override
	public void SetPosition(int X, int Y)
	{
		LPBtn.x = X;
		LPBtn.y = Y;
		OverlayWindow.WM.updateViewLayout(Btn.getObject(), LPBtn);
	}

	@Override
	public int getWidth()
	{
		return LPBtn.width;
	}
	@Override
	public void setWidth(int Width)
	{
		LPBtn.width = Width;
		OverlayWindow.WM.updateViewLayout(Btn.getObject(), LPBtn);
	}

	@Override
	public int getHeight()
	{
		return LPBtn.height;
	}
	@Override
	public void setHeight(int Height)
	{
		LPBtn.height = Height;
		OverlayWindow.WM.updateViewLayout(Btn.getObject(), LPBtn);
	}

	@Override
	public void SetSize(int Width, int Height)
	{
		LPBtn.width = Width;
		LPBtn.height = Height;
		OverlayWindow.WM.updateViewLayout(Btn.getObject(), LPBtn);
	}
}
