package b4a.flm.overlaywdw;

import android.os.Build;
import android.view.View;
import android.view.Gravity;
import android.view.MotionEvent;
import android.view.GestureDetector;
import android.view.View.OnTouchListener;
import android.view.WindowManager;
import android.view.WindowManager.LayoutParams;
import android.content.Context;
import android.graphics.PixelFormat;
import java.util.ArrayList;
import anywheresoftware.b4a.BA;
import anywheresoftware.b4a.objects.PanelWrapper;

//import android.util.FloatMath;
//import anywheresoftware.b4a.BA.Events;
//import android.view.View.OnClickListener;
//import android.view.View.OnLongClickListener;

@BA.ShortName("OverlayWindow")
@BA.Version(1.21F)
@BA.Author("F.Leneuf-Magaud/Informatix")
@BA.Permissions(values = { "android.permission.SYSTEM_ALERT_WINDOW", "android.permission.SYSTEM_ERROR_WINDOW" })

@BA.Events(values =
	{
		"onTouch(Action As Int, X As Float, Y As Float, ScreenX As Float, ScreenY As Float, MotionEvent As Object) As Boolean",
		"onPointerDown(ptrIndex As Int, PID As Int, MotionEvent As Object)",
		"onPointerUp(ptrIndex As Int, PID As Int, MotionEvent As Object)",
		"onDown(X As Float, Y As Float, MotionEvent As Object)",
		"onSingleTapUp(X As Float, Y As Float, MotionEvent As Object)",
		"onSingleTapConfirmed(X As Float, Y As Float, MotionEvent As Object)",
		"onDoubleTap(X As Float, Y As Float, MotionEvent As Object)",
		"onShowPress(X As Float, Y As Float, MotionEvent As Object)",
		"onLongPress(X As Float, Y As Float, MotionEvent As Object)",
		"onDrag(DeltaX As Float, DeltaY As Float, ScreenX As Float, ScreenY As Float, MotionEvent As Object)",
		"onScroll(DistanceX As Float, DistanceY As Float, MotionEvent1 As Object, MotionEvent2 As Object)",
		"onFling(VelocityX As Float, VelocityY As Float, MotionEvent1 As Object, MotionEvent2 As Object)",
		"onPinchOpen(NewDistance As Float, PreviousDistance As Float, MotionEvent As Object)",
		"onPinchClose(NewDistance As Float, PreviousDistance As Float, MotionEvent As Object)",
		"onRotation(Degrees As Double, MotionEvent As Object)"
	})
	
public class OverlayWindow implements OverlayObject
{
	static WindowManager WM = (WindowManager) BA.applicationContext.getSystemService(Context.WINDOW_SERVICE);

	PanelWrapper Pnl;
	LayoutParams LPPnl;
	boolean VueAjouter;
	ArrayList<OverlayButton> Bound;
	int _TouchMode;
	int _OverBars;
	int _KeepScreenOn;
	
	private GestureDetector mGestureDetector;
	private BA _ba;
	private String Prefixe;
	private View _view;
	private float previousDistance;
	private double previousAngle;
	private boolean ignoreHeight;
	private boolean ignoreWidth;
	private float StartX, StartY;
	private boolean ResetStartPos;

	public static final int ACTION_DOWN = MotionEvent.ACTION_DOWN;
	public static final int ACTION_UP = MotionEvent.ACTION_UP;
	public static final int ACTION_MOVE = MotionEvent.ACTION_MOVE;
	public static final int ACTION_CANCEL = MotionEvent.ACTION_CANCEL;

	/**
	 *Creates an empty overlay window at the specified position. The window is displayed after calling Open.
	 *The window has a panel which can contain child views. Note that these views cannot acquire the focus.
	 *The window cannot be placed or moved over the status and notification bars.
	 *@param Width Width of the window.
	 *@param Height Height of the window.
	 *@param EventPrefix Prefix of the Touch, Click and LongClick events.
	 */
	public void Initialize(final BA ba, int X, int Y, int Width, int Height, final String EventPrefix)
	{
		InnerInitialize(ba, X, Y, Width, Height, false, EventPrefix);
	}

	/**
	 *Creates an empty overlay window at the specified position. The window is displayed after calling Open.
	 *The window has a panel which can contain child views. Note that these views cannot acquire the focus.
	 *The window can be placed or moved over the status and notification bars. Starting from Android 8 (Oreo), it always appears under critical system windows.
	 *@param Width Width of the window.
	 *@param Height Height of the window.
	 *@param EventPrefix Prefix of the Touch, Click and LongClick events.
	 */
	public void Initialize2(final BA ba, int X, int Y, int Width, int Height, final String EventPrefix)
	{
		InnerInitialize(ba, X, Y, Width, Height, true, EventPrefix);
	}

	@SuppressWarnings("deprecation")
	private void InnerInitialize(final BA ba, int X, int Y, int Width, int Height, boolean OverBars, final String EventPrefix)
	{
		Pnl = new PanelWrapper();
		Pnl.Initialize(ba, "");
		if (EventPrefix.length() > 0)
		{
			_ba = ba;
			//final String Prefixe = EventPrefix.toLowerCase(BA.cul);
			Prefixe = EventPrefix.toLowerCase(BA.cul);
			mGestureDetector = new GestureDetector(ba.context, new GestureListener());
			previousDistance = -1;
			previousAngle = 999;
			ignoreHeight = false;
			ignoreWidth = false;
			
			OnTouchListener TouchListener = new OnTouchListener()
			{
				@Override
				public boolean onTouch(View Sender, MotionEvent event)
				{
					final int action = event.getAction();
					boolean IsHandled = true;
					if (ba.subExists(Prefixe + "_ontouch"))
					{
						Object Result = ba.raiseEvent(Sender, Prefixe + "_ontouch", new Object[] {action, event.getX(), event.getY(), event.getRawX(), event.getRawY(), event});
						if (Result instanceof Boolean) IsHandled = (boolean) Result;
					}
					
					final int actionCode = action & MotionEvent.ACTION_MASK;
					if (actionCode == MotionEvent.ACTION_POINTER_DOWN && ba.subExists(Prefixe + "_onpointerdown")) {
						final int actionPtrIndex = (action & MotionEvent.ACTION_POINTER_ID_MASK) >> MotionEvent.ACTION_POINTER_ID_SHIFT; 
						ba.raiseEvent(Sender, Prefixe + "_onpointerdown", new Object[] {actionPtrIndex, getPID(event, actionPtrIndex), event});
					}
					else if (actionCode == MotionEvent.ACTION_POINTER_UP && ba.subExists(Prefixe + "_onpointerup")) {
						final int actionPtrIndex = (action & MotionEvent.ACTION_POINTER_ID_MASK) >> MotionEvent.ACTION_POINTER_ID_SHIFT;
						ResetStartPos = true;
						ba.raiseEvent(Sender, Prefixe + "_onpointerup", new Object[] {actionPtrIndex, getPID(event, actionPtrIndex), event});
					}
					else if (actionCode == MotionEvent.ACTION_DOWN) {
						ResetStartPos = true;
						if (ba.subExists(Prefixe + "_ondown")){
							ba.raiseEvent(Sender, Prefixe + "_ondown", new Object[] {StartX, StartY, event});
						}
					}
					else if (actionCode == MotionEvent.ACTION_MOVE && ba.subExists(Prefixe + "_ondrag")) {
						if (ResetStartPos)
						{
							StartX = event.getX();
							StartY = event.getY();
							ResetStartPos = false;
						}
						float DeltaX = event.getX() - StartX;
						float DeltaY = event.getY() - StartY;
						ba.raiseEvent(Sender, Prefixe + "_ondrag", new Object[] {DeltaX, DeltaY, event.getRawX(), event.getRawY(), event});
					}
					
					if (event.getPointerCount() > 1) DetectMultiTouchGestures(ba, Sender, event);
					mGestureDetector.onTouchEvent(event);
					return IsHandled;
				}
			};
			Pnl.getObject().setOnTouchListener(TouchListener);
		}
		
		_TouchMode = TOUCH_INSIDE;
		int LP;
		if (OverBars)
		{
			_OverBars = LayoutParams.FLAG_LAYOUT_IN_SCREEN;
			if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O)
				LP = LayoutParams.TYPE_APPLICATION_OVERLAY;
			else
				LP = LayoutParams.TYPE_SYSTEM_ERROR;
			LPPnl = new LayoutParams(Width, Height, X, Y, LP, _TouchMode | _OverBars, PixelFormat.TRANSLUCENT);
		}
		else
		{
			_OverBars = 0;
			if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O)
				LP = LayoutParams.TYPE_APPLICATION_OVERLAY;
			else
				LP = LayoutParams.TYPE_SYSTEM_ALERT;
			LPPnl = new LayoutParams(Width, Height, X, Y, LP, _TouchMode | _OverBars, PixelFormat.TRANSLUCENT);
		}
		LPPnl.gravity = Gravity.TOP + Gravity.LEFT;
		Bound = new ArrayList<OverlayButton>();
		VueAjouter = false;
	}

	private void DetectMultiTouchGestures(final BA ba, final View Sender, MotionEvent event)
	{
		float distance;
		float distX = event.getX(1) - event.getX(0);
		float distY = event.getY(1) - event.getY(0);
		
		if (ignoreHeight) {
			distance = Math.abs(distX);
		} else if (ignoreWidth) {
			distance = Math.abs(distY);
		} else {
			distance = (float)Math.sqrt(distX * distX + distY * distY);
		}
		if (previousDistance != -1)
		{
			if (distance > previousDistance) {
				if (ba.subExists(Prefixe + "_onpinchopen"))
				{
					ba.raiseEvent(Sender, Prefixe + "_onpinchopen", new Object[] {distance, previousDistance, event});
				}
			}
			else if (distance < previousDistance) {
				if (ba.subExists(Prefixe + "_onpinchclose"))
				{
					ba.raiseEvent(Sender, Prefixe + "_onpinchclose", new Object[] {distance, previousDistance, event});
				}
			}
		}
		previousDistance = distance;

		if (ba.subExists(Prefixe + "_onrotation")) {
			double degrees = Math.toDegrees(Math.atan2(distY, distX));
			if (previousAngle != degrees)
			{
				ba.raiseEvent(Sender, Prefixe + "_onrotation", new Object[] {degrees, event});
				previousAngle = degrees;
			}
		}
	}

	private class GestureListener extends GestureDetector.SimpleOnGestureListener
	{
		@Override
		public boolean onSingleTapUp(MotionEvent ev)
		{
			if (_ba.subExists(Prefixe + "_onsingletapup"))
			{
				_ba.raiseEvent(_view, Prefixe + "_onsingletapup", new Object[] {ev.getX(), ev.getY(), ev});
			}
			return true;
		}
		
		@Override
		public boolean onSingleTapConfirmed(MotionEvent ev)
		{
			if (_ba.subExists(Prefixe + "_onsingletapconfirmed"))
			{
				_ba.raiseEvent(_view, Prefixe + "_onsingletapconfirmed", new Object[] {Float.valueOf(ev.getX()), Float.valueOf(ev.getY()), ev});
			}
			return true;
		}
		
		@Override
		public boolean onDoubleTap(MotionEvent ev)
		{
			if (_ba.subExists(Prefixe + "_ondoubletap"))
			{
				_ba.raiseEvent(_view, Prefixe + "_ondoubletap", new Object[] {Float.valueOf(ev.getX()), Float.valueOf(ev.getY()), ev});
			}
			return true;
		}
		
		@Override
		public void onShowPress(MotionEvent ev)
		{
			if (_ba.subExists(Prefixe + "_onshowpress"))
			{
				_ba.raiseEvent(_view, Prefixe + "_onshowpress", new Object[] {Float.valueOf(ev.getX()), Float.valueOf(ev.getY()), ev});
			}
		}
		
		@Override
		public void onLongPress(MotionEvent ev)
		{
			if (_ba.subExists(Prefixe + "_onlongpress"))
			{
				_ba.raiseEvent(_view, Prefixe + "_onlongpress", new Object[] {Float.valueOf(ev.getX()), Float.valueOf(ev.getY()), ev});
			}
		}
		
		@Override
		public boolean onScroll(MotionEvent e1, MotionEvent e2, float distanceX, float distanceY)
		{
			if (_ba.subExists(Prefixe + "_onscroll"))
			{
				_ba.raiseEvent(_view, Prefixe + "_onscroll", new Object[] {distanceX, distanceY, e1, e2});
			}
			return true;
		}
		
		@Override
		public boolean onFling(MotionEvent e1, MotionEvent e2, float velocityX, float velocityY)
		{
			if (_ba.subExists(Prefixe + "_onfling"))
			{
				_ba.raiseEvent(_view, Prefixe + "_onfling", new Object[] {velocityX, velocityY, e1, e2});
			}
			return true;
		}
	}

	
	/**
	 *Returns the PID for the given pointer index.
	 *0 is the first pointer index. 
	 */
	public static int getPID(MotionEvent ev, int ptrIndex)
	{
		return ev.getPointerId(ptrIndex);
	}
	
	/**
	 *Sets whether longpress is enabled. If this is enabled, when a user presses and holds down, you get a longpress event and nothing further. If it's disabled, the user can press and hold down and then later move their finger and you will get scroll events. By default longpress is enabled. 
	 */
	public void EnableLongPress(boolean Enabled)
	{
		mGestureDetector.setIsLongpressEnabled(Enabled);
	}


	@Override
	public boolean IsInitialized()
	{
		return (Pnl != null && Pnl.IsInitialized());
	}

	/** Returns the window panel. */
	public PanelWrapper getPanel()
	{
		return Pnl;
	}

	/** Opens this window. The bound buttons are added, if any. */
	public void Open()
	{
		if (!VueAjouter) {
			OverlayWindow.WM.addView(Pnl.getObject(), LPPnl);
			if (Bound.size() > 0) {
				for (OverlayButton Bouton : Bound) {
					if (Bouton != null)
						Bouton.Add();
				}
			}
			VueAjouter = true;
		}
	}

	/** Closes this window. The bound buttons are removed, if any. */
	public void Close()
	{
		if (VueAjouter) {
			OverlayWindow.WM.removeView(Pnl.getObject());
			if (Bound.size() > 0) {
				for (OverlayButton Bouton : Bound) {
					if (Bouton != null)
						Bouton.Remove();
				}
			}
			VueAjouter = false;
		}
	}

	/** Returns True if the window was opened by a call to Open. */
	public boolean IsOpen()
	{
		return VueAjouter;
	}

	/** Gets or sets the window visibility. The bound buttons, if any, are shown or hidden accordingly. */
	public boolean getVisible()
	{
		return Pnl.getVisible();
	}
	public void setVisible(boolean Visible)
	{
		Pnl.setVisible(Visible);
		if (Bound.size() > 0) {
			for (OverlayButton Bouton : Bound) {
				if (Bouton != null)
					Bouton.getButton().setVisible(Visible);
			}
		}
	}

	@Override
	public int getX()
	{
		return LPPnl.x;
	}
	@Override
	public void setX(int X)
	{
		LPPnl.x = X;
		OverlayWindow.WM.updateViewLayout(Pnl.getObject(), LPPnl);
	}

	@Override
	public int getY()
	{
		return LPPnl.y;
	}
	@Override
	public void setY(int Y)
	{
		LPPnl.y = Y;
		OverlayWindow.WM.updateViewLayout(Pnl.getObject(), LPPnl);
	}

	@Override
	public void SetPosition(int X, int Y)
	{
		LPPnl.x = X;
		LPPnl.y = Y;
		OverlayWindow.WM.updateViewLayout(Pnl.getObject(), LPPnl);
	}

	@Override
	public int getWidth()
	{
		return LPPnl.width;
	}
	@Override
	public void setWidth(int Width)
	{
		LPPnl.width = Width;
		OverlayWindow.WM.updateViewLayout(Pnl.getObject(), LPPnl);
	}

	@Override
	public int getHeight()
	{
		return LPPnl.height;
	}
	@Override
	public void setHeight(int Height)
	{
		LPPnl.height = Height;
		OverlayWindow.WM.updateViewLayout(Pnl.getObject(), LPPnl);
	}

	@Override
	public void SetSize(int Width, int Height)
	{
		LPPnl.width = Width;
		LPPnl.height = Height;
		OverlayWindow.WM.updateViewLayout(Pnl.getObject(), LPPnl);
	}

	private void SetAllFlags()
	{
		LPPnl.flags = _TouchMode | _OverBars | _KeepScreenOn;
		OverlayWindow.WM.updateViewLayout(Pnl.getObject(), LPPnl);
	}

	/** The window does not receive touch events. */
	public final static int TOUCH_NONE = LayoutParams.FLAG_NOT_FOCUSABLE | LayoutParams.FLAG_NOT_TOUCHABLE;
	/** The window receives touch events inside the window. */
	public final static int TOUCH_INSIDE = LayoutParams.FLAG_NOT_FOCUSABLE;
	/** The window receives all touch events. */
	public final static int TOUCH_ALL = LayoutParams.FLAG_NOT_FOCUSABLE | LayoutParams.FLAG_NOT_TOUCH_MODAL | LayoutParams.FLAG_WATCH_OUTSIDE_TOUCH;

	/** Gets or sets the touch mode (one of the TOUCH_ constants). Default = TOUCH_INSIDE. Should be set only after the window is opened. */
	public int getTouchMode()
	{
		return _TouchMode;
	}
	public void setTouchMode(int TouchMode)
	{
		_TouchMode = TouchMode;
		SetAllFlags();
	}

	/** Gets or sets whether the screen is kept turned on and bright as long as the window is visible. Default = False. Should be set only after the window is opened. */
	public boolean getKeepScreenOn()
	{
		return (_KeepScreenOn == LayoutParams.FLAG_KEEP_SCREEN_ON);
	}
	public void setKeepScreenOn(boolean KeepScreenOn)
	{
		if (KeepScreenOn)
			_KeepScreenOn = LayoutParams.FLAG_KEEP_SCREEN_ON;
		else
			_KeepScreenOn = 0;
		SetAllFlags();
	}

}
